import os
import bandwitch.tools as tools

this_directory = os.path.dirname(os.path.realpath(__file__))


def test_set_record_topology():
    record_path = os.path.join(
        this_directory, "test_data", "test_sequences", "asm_00.gb"
    )
    record = tools.load_record(record_path, topology="circular")
    assert record.annotations["topology"] == "circular"
    assert not tools.record_is_linear(record, default=True)

    # Default only when no topology is set
    tools.set_record_topology(record, "default_to_circular")
    assert not tools.record_is_linear(record, default=True)
    record.annotations.pop("topology")
    assert not tools.record_is_linear(record, default=False)
    tools.set_record_topology(record, "default_to_linear")
    assert tools.record_is_linear(record, default=False)



"""Classes for parsing and analysis of DNA migration data."""

from collections import OrderedDict, Counter
import matplotlib.pyplot as plt
from io import BytesIO
import pandas
from matplotlib.backends.backend_pdf import PdfPages

from bandwagon import BandsPattern, BandsPatternsSet
import flametree

try:
    from plateo.exporters.plate_to_matplotlib_plots import PlateColorsPlotter
    from plateo.parsers import plate_from_platemap_spreadsheet
    from plateo.containers import Plate96

    PLATEO_AVAILABLE = True
except ImportError:
    PLATEO_AVAILABLE = False

try:
    import saboteurs

    SABOTEURS_AVAILABLE = True
except ImportError:
    SABOTEURS_AVAILABLE = False

from ..bands_predictions import predict_digestion_bands
from ..tools import load_record, all_subsets, record_is_linear
from .Clone import Clone
from .BandsObservation import BandsObservation


class ClonesObservations:
    """All useful information for a collection of clones to be validated.

    Parameters
    ----------
    clones : list or dict
        Either a list of Clones (each with a unique name) or a dictionary
        ``{name: Clone}``.
    constructs_records : dict
        A dictionary ``{construct_name: biopython_record}`` indicating the
        sequence of the different constructs. For each construct, set the
        attribute ``construct.linear = False`` if the construct is circular.
    """

    def __init__(self, clones, constructs_records, partial_cutters=()):
        """Initialize."""
        if isinstance(clones, (list, tuple)):
            clones = OrderedDict([(clone.name, clone) for clone in clones])
        self.clones = clones
        clones_constructs = []
        for clone in list(clones.values()):
            if clone.construct_id not in clones_constructs:
                clones_constructs.append(clone.construct_id)
        self.constructs_records = OrderedDict(
            sorted(
                [
                    (construct_name, record)
                    for construct_name, record in constructs_records.items()
                ],
                key=lambda cst: (
                    +100000
                    if cst[0] not in clones_constructs
                    else clones_constructs.index(cst[0])
                ),
            )
        )
        self.constructs_digestions = {ct: {} for ct in constructs_records}
        self.partial_cutters = partial_cutters

    def get_digestion_bands_for_construct(self, construct_id, digestion):
        """Return the bands resulting from the digestion of the construct.

        This function enables some memoization (no digestion is computed
        twice).

        Parameters
        ----------
        construct_id : str
            ID of the construct as it appears in this object's ``constructs_records``.
        digestion : tuple
            For instance ``('BamHI', 'XbaI')``.
        """
        construct_digestions = self.constructs_digestions[construct_id]
        if digestion not in construct_digestions:
            construct_record = self.constructs_records[construct_id]
            construct_digestions[digestion] = predict_digestion_bands(
                str(construct_record.seq),
                linear=record_is_linear(construct_record, default=True),
                enzymes=[
                    enzyme for enzyme in digestion if enzyme not in self.partial_cutters
                ],
                partial_cutters=self.partial_cutters,
            )
        return construct_digestions[digestion]

    def get_clone_digestion_bands(self, clone, construct_id=None):
        """Return ``{digestion: bands}`` for all digestions of this clone."""
        if construct_id is None:
            construct_id = clone.construct_id
        return {
            digestion: self.get_digestion_bands_for_construct(
                construct_id=clone.construct_id, digestion=digestion
            )
            for digestion in clone.digestions
        }

    def validate_all_clones(
        self,
        relative_tolerance=0.05,
        min_band_cutoff=None,
        max_band_cutoff=None,
    ):
        """Return ``{clone: CloneValidation}`` for all clones."""
        return OrderedDict(
            [
                (
                    clone_name,
                    clone.validate_bands(
                        bands_by_digestion=self.get_clone_digestion_bands(clone),
                        relative_tolerance=relative_tolerance,
                        min_band_cutoff=min_band_cutoff,
                        max_band_cutoff=max_band_cutoff,
                    ),
                )
                for clone_name, clone in self.clones.items()
            ]
        )

    def identify_all_clones(
        self,
        relative_tolerance=0.05,
        min_band_cutoff=None,
        max_band_cutoff=None,
    ):
        """Return ``{clone: {construct_id: CloneValidation}}`` for all clones.

        Parameters
        ----------
        relative_tolerance : float
            Tolerance, as a ratio of the full ladder span. If =0.1, then the
            discrepancy will have a value of 1 when a band's nearest
            correspondent in the other pattern is more that 10% of the ladder
            span apart.
        min_band_cutoff : int
            Discrepancies involving at least one band below this minimal band
            size will be ignored. By default, it will be set to the smallest
            band size in the ladder.
        max_band_cutoff : int
            Discrepancies involving at least one band above this minimal band
            size will be ignored. By default, it will be set to the smallest
            band size in the ladder.
        """
        return OrderedDict(
            [
                (
                    clone_name,
                    {
                        construct_id: clone.validate_bands(
                            bands_by_digestion=self.get_clone_digestion_bands(
                                clone, construct_id=construct_id
                            ),
                            relative_tolerance=relative_tolerance,
                            min_band_cutoff=min_band_cutoff,
                            max_band_cutoff=max_band_cutoff,
                        )
                        for construct_id in self.constructs_records
                    },
                )
                for clone_name, clone in self.clones.items()
            ]
        )

    def validations_summary(self, validations, sort_clones_by_score=True):
        """Return ``{construct_id: [CloneValidation, ...]}``.

        To each construct corresponds a list of the validation of all clones
        associated with that construct, from the best-scoring to the
        least-scoring.
        """
        results = OrderedDict()
        for clone_name, validation in validations.items():
            construct_id = self.clones[clone_name].construct_id
            if construct_id not in results:
                results[construct_id] = []
            results[construct_id].append(validation)
        if sort_clones_by_score:
            for cst_id, validations in results.items():
                results[cst_id] = sorted(validations, key=lambda v: v.max_discrepancy)
        return results

    def plot_validations_plate_map(self, validations, target=None, ax=None):
        """Plot a map of the plate with passing/failing wells in green/red."""

        if not PLATEO_AVAILABLE:
            raise ImportError("This function requires Plateo installed")
        plate = Plate96("Validations map")
        constructs_colors = {}
        for well in plate.iter_wells():
            if well.name not in validations:
                well.data["bg_color"] = None
                well.data["pass_color"] = (1, 1, 1, 0.1)
            else:
                validation = validations[well.name]
                construct = validation.clone.construct_id
                if construct not in constructs_colors:
                    alpha = ((0.35 * len(constructs_colors) % 1) + 0.2) / 2
                    constructs_colors[construct] = (0.4, 0.4, 0.76, alpha)
                well.data["bg_color"] = constructs_colors[construct]
                well.data["pass_color"] = (
                    (0.43, 0.92, 0.49) if validation.passes else (0.91, 0.43, 0.43)
                )
        bg_plotter = PlateColorsPlotter(lambda w: w.data["bg_color"])
        ax, _ = bg_plotter.plot_plate(plate, figsize=(10, 5), ax=ax)
        pass_plotter = PlateColorsPlotter(
            lambda w: w.data["pass_color"], well_radius=250
        )
        pass_plotter.plot_plate(plate, ax=ax)
        if target is not None:
            ax.figure.savefig(target, bbox_inches="tight")
            plt.close(ax.figure)
        else:
            return ax

    def validations_summary_table(self, validations, target=None):
        validations_summary = self.validations_summary(validations)
        records = []
        for construct, data in validations_summary.items():
            best_clone = None
            valid_clones = [clone for clone in data if clone.passes]
            best_clone_name = None
            if len(valid_clones):
                best_clone = min(valid_clones, key=lambda c: c.max_discrepancy)
                best_clone_name = best_clone.clone.name
            other_valid_clones = ", ".join(
                [
                    clone.clone.name
                    for clone in valid_clones
                    if clone.clone.name != best_clone_name
                ]
            )
            records.append(
                dict(
                    construct=construct,
                    n_clones=len(data),
                    valid_clones=len(valid_clones),
                    best_clone=best_clone_name,
                    other_valid_clones=other_valid_clones,
                )
            )
        dataframe = pandas.DataFrame.from_records(
            records,
            columns=[
                "construct",
                "n_clones",
                "valid_clones",
                "best_clone",
                "other_valid_clones",
            ],
        )
        if target is not None:
            dataframe.to_csv(target, index=False)
        return dataframe

    def plot_all_validations_patterns(
        self, validations, target=None, per_digestion_discrepancy=False
    ):
        """Plot a Graphic report of the gel validation.

        The report displays the patterns, with green and red backgrounds
        depending on whether they passed the validation test.

        Parameters
        ----------
        validations
            Validation results.
        target
            File path.
        per_digestion_discrepancy
            If true, each pattern for each clone will have an indication of
            how close it is from the expected pattern. If False, and the
            validation involves several digestions, the biggest discrepancy
            among all patterns is shown.
        """

        summary = self.validations_summary(validations)
        max_x = (
            Counter([clone.construct_id for clone in self.clones.values()]).most_common(
                1
            )[0][1]
            + 2
        )
        digestions_by_construct = {construct: [] for construct in summary}
        for construct, validations in summary.items():
            for validation in validations:
                for digestion in validation.discrepancies:
                    if digestion not in digestions_by_construct[construct]:
                        digestions_by_construct[construct].append(digestion)

        # ladder = list(validations[0].clone.digestions.values())[0]["ladder"]
        pdf_io = BytesIO()
        with PdfPages(pdf_io) as pdf:
            for construct_id, validations in summary.items():
                ladder = list(validations[0].clone.digestions.values())[0].ladder
                digestions = digestions_by_construct[construct_id]
                band_patterns = OrderedDict()
                seen_clones = set()
                for i, digestion in enumerate(digestions):
                    reference_bands = self.get_digestion_bands_for_construct(
                        construct_id, digestion
                    )
                    reference = BandsPattern(
                        bands=reference_bands,
                        ladder=ladder,
                        label="exp." if (i == 0) else None,
                        background_color="#c6dcff",
                        corner_note="Total: %d bp" % sum(reference_bands),
                        global_bands_props={"label_fontdict": {"size": 5}},
                    )
                    sorted_bands = sorted(reference.bands, key=lambda b: -b.dna_size)
                    for band_name, band in zip("abcdefghijklm", sorted_bands):
                        band.label = band_name

                    patterns = []
                    for validation in validations:
                        clone_name = validation.clone.name
                        label = None if clone_name in seen_clones else "auto"
                        pattern = validation.to_bandwagon_bandpattern(
                            digestion=digestion,
                            label=label,
                            per_digestion_discrepancy=per_digestion_discrepancy,
                        )
                        if pattern is not None:
                            seen_clones.add(clone_name)
                        patterns.append(pattern)
                    band_patterns[digestion] = BandsPatternsSet(
                        [reference] + patterns,
                        ladder=ladder,
                        label=" + ".join(digestion),
                        ladder_ticks=5,
                        global_patterns_props={"label_fontdict": {"rotation": 60}},
                    )

                digestions = digestions_by_construct[construct_id]
                fig, axes = plt.subplots(
                    len(digestions),
                    1,
                    sharex=True,
                    figsize=(1.1 * max_x, 3 * len(digestions)),
                )
                if len(digestions) == 1:
                    axes = [axes]
                axes[-1].set_xlim(right=max_x)

                for ax, (dig, pattern_set) in zip(axes, band_patterns.items()):
                    pattern_set.plot(ax)

                axes[-1].set_xlabel(
                    construct_id,
                    fontdict=dict(size=14, weight="bold"),
                    ha="left",
                    position=(0.0, 0.1),
                )
                pdf.attach_note(construct_id)

                # Temporary fix because of Matplotlib bug #12634
                pdf.savefig(fig, bbox_inches="tight")
                plt.close(fig)
        pdf_data = pdf_io.getvalue()
        if target is None:
            return pdf_data
        elif target == "base64":
            return "data:application/pdf;base64," + pdf_data.decode("utf-8")
        else:
            with open(target, "wb") as f:
                f.write(pdf_data)

    @staticmethod
    def identify_bad_parts(
        validations,
        constructs_parts,
        constructs_records=None,
        report_target=None,
        extra_failures=None,
    ):
        """Identifies parts associated with failure in the validations.

        Uses the Saboteurs package:
        https://github.com/Edinburgh-Genome-Foundry/saboteurs

        Parameters
        ----------
        validations
            Validation results.

        constructs_parts
            Either a dict {construct_id: [list, of, part, names]} or a function
            (biopython_record => [list, of, part, names]).

        report_target
            Can be a path to file or file-like object where to write a PDF
            report. Can also be "@memory", at which case the raw binary PDF
            data is returned.

        Returns
        -------
        analysis, pdf_data
            Where ``analysis`` is the result of sabotage analysis (see the
            saboteurs library), and pdf_data is None unless report_target is
            set to "@memory" (see above).
        """
        if not SABOTEURS_AVAILABLE:
            raise ImportError(
                "You must install the saboteurs library to"
                "be able to identify bad parts. Try:\n\n"
                "(sudo) pip install saboteurs"
            )
        if hasattr(constructs_parts, "__call__"):
            constructs_parts = {
                record_id: constructs_parts(record)
                for record_id, record in constructs_records.items()
            }
        extra_failures = extra_failures or {}
        constructs_stats = OrderedDict()
        for clone_name, validation in validations.items():
            construct_id = validation.clone.construct_id
            if construct_id not in constructs_stats:
                constructs_stats[construct_id] = dict(
                    exp_id=construct_id,
                    attempts=0,
                    failures=0,
                    members=constructs_parts[construct_id],
                )
            stats = constructs_stats[construct_id]
            stats["attempts"] += 1
            stats["failures"] += not validation.passes
        for construct_id, weight in extra_failures.items():
            constructs_stats[construct_id] = dict(
                exp_id=construct_id,
                attempts=weight,
                failures=weight,
                members=constructs_parts[construct_id],
            )
        # import json
        # print (json.dumps(constructs_stats, indent=2))
        analysis = saboteurs.find_statistical_saboteurs(constructs_stats)
        report_data = None
        if report_target is not None:
            report_data = saboteurs.statistics_report(analysis, report_target)
        return analysis, report_data

    def write_identification_report(
        self,
        target_file=None,
        relative_tolerance=0.05,
        min_band_cutoff=None,
        max_band_cutoff=None,
    ):
        """Plot a Graphic report of the gel validation.

        The report displays the patterns, with green and red backgrounds
        depending on whether they passed the validation test.

        Parameters
        ----------
        target_file
            File object or file path where to save the figure. If provided, the
            function returns none, else the function returns the axes array of
            the final figure.
        relative_tolerance
            Relative error tolerated on each band for the ovserved patterns to
            be considered similar to the expected patterns.
        min_band_cutoff
            Bands with a size below this value will not be considered.
        max_band_cutoff
            Bands with a size above this value will not be considered.
        """
        bands_validities = self.compute_all_bands_validities(
            relative_tolerance=relative_tolerance,
            min_band_cutoff=min_band_cutoff,
            max_band_cutoff=max_band_cutoff,
        )
        L = len(self.constructs_records)
        max_x = max(len(measures) for measures in self.observed_bands.values()) + 1
        fig, axes = plt.subplots(L, 1, figsize=(2.2 * max_x, 3 * L))
        axes_validities = zip(axes, bands_validities.items())
        for ax, (construct_id, validities) in axes_validities:
            reference = BandsPattern(
                self.expected_bands[construct_id],
                ladder=self.ladder,
                label="exp.",
                background_color="#c6dcff",
                corner_note="Total: %d bp" % sum(self.expected_bands[construct_id]),
                global_bands_props={"label_fontdict": {"size": 5}},
            )
            sorted_bands = sorted(reference.bands, key=lambda b: -b.dna_size)
            for band_name, band in zip("abcdefghijklm", sorted_bands):
                band.label = band_name
            patterns = [
                BandsPattern(
                    self.observed_bands[construct_id][measure_name],
                    corner_note="Total: %d bp"
                    % sum(self.observed_bands[construct_id][measure_name]),
                    ladder=self.ladder,
                    label=measure_name,
                    gel_image=self.migration_images[measure_name],
                    background_color=(
                        "#aaffaa" if validities[measure_name] else "#ffaaaa"
                    ),
                )
                for measure_name in self.observed_bands[construct_id]
            ]
            patterns_set = BandsPatternsSet(
                [reference] + patterns,
                ladder=self.ladder,
                label=construct_id,
                ladder_ticks=5,
                global_patterns_props={"label_fontdict": {"rotation": 60}},
            )
            ax.set_xlim(0.5, max_x + 2)
            patterns_set.plot(ax)
        fig.subplots_adjust(hspace=0.3)
        if target_file is not None:
            fig.savefig(target_file, bbox_inches="tight")
            plt.close(fig)
        else:
            return axes

    @staticmethod
    def from_files(
        records_path,
        constructs_map_path,
        aati_zip_path,
        digestions_map_path=None,
        digestion=None,
        direction="column",
        ignore_bands_under=None,
        topology="circular",
    ):
        constructs_records = {
            f._name_no_extension: load_record(
                f.open("r"),
                topology=topology,
                id=f._name_no_extension,
                file_format="genbank",
            )
            for f in flametree.file_tree(records_path)._all_files
        }
        constructs_records = OrderedDict(sorted(constructs_records.items()))

        constructs_plate = plate_from_platemap_spreadsheet(
            constructs_map_path, data_field="construct", headers=True
        )
        constructs_map = OrderedDict(
            [
                (well.name, well.data["construct"])
                for well in constructs_plate.iter_wells(direction=direction)
                if "construct" in well.data and str(well.data["construct"]) != "nan"
            ]
        )

        if digestion:
            digestion = tuple(digestion)
            digestions_map = OrderedDict(
                [(wellname, digestion) for wellname in constructs_map]
            )
        else:
            digestions_plate = plate_from_platemap_spreadsheet(
                digestions_map_path, data_field="digestion", headers=True
            )
            digestions_map = OrderedDict(
                [
                    (well.name, tuple(well.data["digestion"].split(", ")))
                    for well in digestions_plate.iter_wells(direction=direction)
                    if "digestion" in well.data and str(well.data["digestion"]) != "nan"
                ]
            )

        observations = BandsObservation.from_aati_fa_archive(
            aati_zip_path,
            direction=direction,
            ignore_bands_under=ignore_bands_under,
        )
        clones = Clone.from_bands_observations(
            observations, constructs_map, digestions_map
        )
        clones_observations = ClonesObservations(clones, constructs_records)
        return clones_observations

    def partial_digests_analysis(self, relative_tolerance=0.05):
        """Compute good clones under different partial digest assumptions.

        Returns a dictionary ``{partial: {'valid_clones': 60, 'label': 'x'}}``
        where for a given scenario ``partial`` is a tuple of all enzymes
        considered to have partial activity in this scenario (it defines the
        scenario) ``valid_clones`` is the number of good clones under this
        assumption, and ``label`` is a string representation of all enzymes
        involved, with partial activity enzymes in parenthesis.

        This result can be fed to ``ClonesObservations``'s
        ``.plot_partial_digests_analysis`` method for plotting.
        """
        all_enzymes = set(
            enzyme
            for clone in self.clones.values()
            for digestion in clone.digestions
            for enzyme in digestion
        )
        all_enzymes
        results = {}
        for good_cutters in all_subsets(all_enzymes):
            partial_cutters = tuple([e for e in all_enzymes if e not in good_cutters])
            clones_observations = ClonesObservations(
                self.clones,
                self.constructs_records,
                partial_cutters=partial_cutters,
            )
            validations = clones_observations.validate_all_clones(
                relative_tolerance=relative_tolerance
            )
            label = " + ".join(
                sorted(good_cutters) + sorted(["(%s)" % c for c in partial_cutters])
            )
            valid_clones = sum([v.passes for name, v in validations.items()])

            results[partial_cutters] = {
                "label": label,
                "validations": validations,
                "valid_clones": valid_clones,
            }
        return results

    @staticmethod
    def plot_partial_digests_analysis(analysis_results, ax=None):
        """Plot partial digest analysis results.

        Parameters
        ----------
        analysis_results
            results from ``ClonesObservations.partial_digest_analysis``.
        ax
            A Matplotlib ax. If none, one is created and returned at the end.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(5, 0.5 * len(analysis_results)))
        ax.axis("off")
        values, labels = zip(
            *sorted(
                [
                    (data["valid_clones"], data["label"])
                    for partial, data in analysis_results.items()
                ]
            )
        )
        ax.barh(range(len(values)), values)
        for i, (value, label) in enumerate(zip(values, labels)):
            ax.text(
                0,
                i,
                label + " ",
                ha="right",
                va="center",
                fontweight=("normal" if "(" in label else "bold"),
            )
            ax.text(
                value,
                i,
                str(value) + " ",
                ha="right",
                va="center",
                fontdict={"color": "white", "weight": "bold"},
            )
        ax.set_ylim(-0.5, len(values) - 0.5)
        ax.set_title("Number of good clones, by scenario")
        return ax

    @staticmethod
    def merge_observations(observations_list, prefixes=None):
        prefixes = prefixes or (len(observations_list) * [""])

        clones = OrderedDict(
            [
                (pref + k, v)
                for pref, obs in zip(prefixes, observations_list)
                for k, v in obs.clones.items()
            ]
        )
        constructs_records = {
            name: record
            for obs in observations_list
            for name, record in obs.constructs_records.items()
        }
        partial_cutters = tuple(
            enzyme for obs in observations_list for enzyme in obs.partial_cutters
        )

        return ClonesObservations(clones, constructs_records, partial_cutters)



from collections import OrderedDict
from .CloneValidation import CloneValidation


class Clone:
    """Gather all information necessary to validate a clone.

    Parameters
    ----------
    name : str
        Name of the clone. Could be for instance a microplate well.
    digestions : dict
        A dictionary ``{digestion: CloneObservation}`` where ``digestion`` is
        of the form ``('EcoRI', 'BamHI')``.
    construct_id : str
        ID of the construct to be validated. This is used to group clones
        by construct in the validation reports.
    """

    def __init__(self, name, digestions, construct_id=None):
        """Initialize."""
        self.name = name
        self.digestions = digestions
        self.construct_id = construct_id

    def validate_bands(
        self,
        bands_by_digestion,
        relative_tolerance=0.1,
        min_band_cutoff=None,
        max_band_cutoff=None,
    ):
        """Return a validation result (comparison of observed and expected).

        The result is a CloneValidation object.

        Parameters
        ----------
        bands_by_digestion : dict
            A dictionary ``{digestion: [bands]}`` where ``digestion`` is
            of the form ``('EcoRI', 'BamHI')``, and bands is a list of band sizes.
        relative_tolerance : float
            Tolerance, as a ratio of the full ladder span. If =0.1, then the
            discrepancy will have a value of 1 when a band's nearest
            correspondent in the other pattern is more than 10% of the ladder
            span apart.
        min_band_cutoff : int, optional
            Discrepancies involving at least one band below this minimal band
            size will be ignored. By default, it will be set to the smallest
            band size in the ladder.
        max_band_cutoff : int, optional
            Discrepancies involving at least one band above this minimum band
            size will be ignored. By default, it will be corrected to the largest
            band size in the ladder.
        """
        for digestion, observation in self.digestions.items():
            discrepancies = {
                digestion: observation.patterns_discrepancy(
                    other_bands=bands_by_digestion[digestion],
                    relative_tolerance=relative_tolerance,
                    min_band_cutoff=min_band_cutoff,
                    max_band_cutoff=max_band_cutoff,
                )
                for digestion, observation in self.digestions.items()
            }
        return CloneValidation(self, bands_by_digestion, discrepancies=discrepancies)

    @staticmethod
    def from_bands_observations(
        observations, constructs_map, digestions_map, clones_map=None
    ):
        if isinstance(observations, (dict, OrderedDict)):
            observations = list(observations.values())
        if clones_map is None:
            clones_map = {name: name for name in constructs_map}
        clones = OrderedDict()

        for obs in observations:
            if obs.name not in clones_map:
                continue
            clone_name = clones_map[obs.name]
            if clone_name not in clones:
                clones[clone_name] = Clone(
                    name=clone_name,
                    construct_id=constructs_map.get(clone_name, "no_clone"),
                    digestions={},
                )
            digestion = digestions_map[obs.name]
            clones[clone_name].digestions[digestion] = obs

        return clones



"""Bandwitch, a Python library for computer-aided restriction digests."""

# __all__ = []

from .DigestionProblem import (
    SeparatingDigestionsProblem,
    IdealDigestionsProblem,
)
from .ClonesObservations import BandsObservation, Clone, ClonesObservations
from .Ladder import Ladder, LADDERS

from .tools import load_record
from .bands_predictions import predict_digestion_bands
from .list_common_enzymes import list_common_enzymes
from .version import __version__


__all__ = [
    "SeparatingDigestionsProblem",
    "IdealDigestionsProblem",
    "BandsObservation",
    "Clone",
    "ClonesObservations",
    "Ladder",
    "LADDERS",
    "load_record",
    "predict_digestion_bands",
    "list_common_enzymes",
    "__version__",
]



"""
Basic tests to check that the core functionalities are at least running.
"""

import matplotlib

matplotlib.use("Agg")

import os
from collections import OrderedDict
from bandwitch import (
    SeparatingDigestionsProblem,
    IdealDigestionsProblem,
    LADDERS,
    list_common_enzymes,
)
from Bio import SeqIO, Restriction
import pytest


@pytest.fixture
def sequences():
    data_path = os.path.join("tests", "test_data", "test_sequences")
    return OrderedDict(
        [
            (
                fname,
                str(SeqIO.read(os.path.join(data_path, fname), "genbank").seq),
            )
            for fname in sorted(os.listdir(data_path))
        ]
    )


def test_separating_digestions(tmpdir, sequences):
    enzymes = sorted(
        [
            str(enzyme)
            for enzyme in Restriction.CommOnly
            if (enzyme.size == 6) and (len(enzyme.supplier_list()) >= 3)
        ]
    )
    problem = SeparatingDigestionsProblem(
        sequences=sequences,
        enzymes=enzymes,
        topology="circular",
        ladder=LADDERS["100_to_4k"],
        max_enzymes_per_digestion=1,
        relative_migration_precision=0.05,
    )
    score, selected_digestions = problem.select_digestions(max_digestions=1)
    assert selected_digestions == [("AvaI",)]

    axes = problem.plot_digestions(
        selected_digestions,
        patterns_props={"label_fontdict": {"rotation": 35}},
    )
    axes[0].figure.savefig(
        os.path.join(str(tmpdir), "separating_digestions.png"),
        bbox_inches="tight",
    )
    ax = problem.plot_distances_map(digestions=selected_digestions)
    ax.figure.savefig(os.path.join(str(tmpdir), "distances.png"))


def test_ideal_digestions(sequences):
    class MyIdealDigestionsProblem(IdealDigestionsProblem):
        def migration_pattern_is_ideal(self, migration):
            """2-3 bands between 30% and 70% of the migration span?"""
            min_migr = 0.7 * self.migration_min + 0.3 * self.migration_max
            max_migr = 0.3 * self.migration_min + 0.7 * self.migration_max
            bands_in_central_zone = [
                band for band in migration if min_migr <= band <= max_migr
            ]
            return 2 <= len(bands_in_central_zone) <= 3

    # DEFINE THE SEQUENCES AND THE ENZYME SET

    enzymes = [
        "EcoRI",
        "BamHI",
        "XhoI",
        "EcoRV",
        "SpeI",
        "XbaI",
        "NotI",
        "SacI",
        "SmaI",
        "HindIII",
        "PstI",
    ]

    # DEFINE AND SOLVE THE PROBLEM

    problem = MyIdealDigestionsProblem(
        sequences=sequences,
        enzymes=enzymes,
        ladder=LADDERS["100_to_4k"],
        topology="circular",
        max_enzymes_per_digestion=2,
    )
    score, selected_digestions = problem.select_digestions(max_digestions=1)
    assert score > 0.09
    assert selected_digestions == [("BamHI", "SacI")]


def test_list_common_enzymes():
    l = list_common_enzymes(min_suppliers=4, avoided_methylations=("Dam", "Dcm"))
    assert len(l) == 38
    l = list_common_enzymes(min_suppliers=4, avoided_methylations=())
    assert len(l) == 45



import numpy as np
from ..tools import max_min_distance


def band_patterns_discrepancy(
    bands1,
    bands2,
    ladder,
    relative_tolerance=0.1,
    reference_and_gel=False,
    zone=None,
):
    """Return a discrepancy indicating whether the band patterns are
    perfectly matching (0) or dissimilar (>=1) or in-between (0-1).

    Similarity is determined by the fact that the min-max distance between
    the two migration patterns is below some threshold. The threshold is some
    proportion of the ladder's migration span.

    If mode ``reference_and_gel`` is set to True, then ``bands1`` is considered
    a "truth" (= expected bands pattern) and the function will automatically
    return False if the observed pattern on gel (bands2) has more bands than
    bands1.

    Parameters
    ----------
    bands1 : list
        A list of bands sizes forming the first pattern.
    bands2 : list
        A list of bands sizes forming the second pattern.
    ladder : Ladder
        A Ladder object used to compute migration distances.
    relative_tolerance : float
        Proportion of the ladder's migration span (between lowest and highest
        migration distances) that is the threshold for the min-max distance.
    reference_and_gel : bool
        Set to True if ``bands1`` is a reference (= expected bands) and
        ``bands2`` is the observed bands.
    zone
    """
    if bands1 == bands2 == []:
        return 0
    elif min(len(bands1), len(bands2)) == 0:
        return 2.0
    if reference_and_gel and (len(bands2) > len(bands1)):
        return 2.0

    m1, m2 = (ladder.dna_size_to_migration(np.array(b)) for b in (bands1, bands2))
    mini, maxi = ladder.migration_distances_span
    tolerance = relative_tolerance * (maxi - mini)
    zone = [ladder.dna_size_to_migration(b) for b in zone][::-1]
    return 1.0 * max_min_distance(m1, m2, zone=zone) / tolerance



import os
import pandas
from Bio import Restriction

this_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))
csv_path = os.path.join(this_dir, "data", "enzymes_infos.csv")
dataframe = pandas.read_csv(csv_path, index_col="enzyme", sep=";")
enzymes_infos = dataframe.T.to_dict()


def list_common_enzymes(
    site_length=(6,),
    opt_temp=(37,),
    min_suppliers=1,
    uniquify_sites=True,
    avoided_methylations=("Dam", "Dcm"),
    accepted_methylation_effects=("-", "cut"),
):
    """List enzymes with many suppliers, no star activity, etc.

    Parameters
    ----------
    site_length
        Tuple of the accepted lengths of enzymes sites, e.g. (6, 4) for 6bp and 4bp.
    opt_temp
        Tuple of the accepted optimal temperatures. You would better keep it
        to a single value though, to ensure that all enzymes returned can
        co-digest.
    min_suppliers
        Minimum number of suppliers. Set to 3 or 4 to get enzymes that should
        be largely commercially available.
    uniquify_sites
        If True and two enzymes in the list have the same recognition site,
        only one will be kept. This speeds up computations in BandWitch in
        practice.
    avoided_methylations
        Tuple of the methylation types to avoid, of Dam, Dcm, CpG... see the
        next parameter for the tolerance level.
    accepted_methylation_effects
        Accepted effects of the methylations specified by "avoided_methylations".
        This data comes from the REBASE online database.
        The effects are among "-" (no overlap, no effect), "cut" (overlap with
        a methylation site, but the enzyme still cuts), "variable" (whatever
        that means), "impaired", "some blocked", "some impaired", and "blocked".
    """
    result = [
        str(e)
        for e in Restriction.AllEnzymes
        if str(e) in enzymes_infos
        and (len(e.site) in site_length)
        and ((opt_temp is None) or (e.opt_temp == 37))
        and enzymes_infos[str(e)]["suppliers"] >= min_suppliers
        and not any(
            [
                enzymes_infos[str(e)][meth] not in accepted_methylation_effects
                for meth in avoided_methylations
            ]
        )
    ]
    if uniquify_sites:
        sites_dict = {}
        for e in result:
            enzyme = Restriction.__dict__[e]
            sites_dict[enzyme.site] = e
        result = list(sites_dict.values())
    return sorted(result)



"""Minimal implementation of a Ladder, which is used to predict band migration
patterns from band DNA sizes."""

import numpy as np
from scipy.interpolate import CubicSpline


class Ladder:
    """Class to represent gel ladders. These ladders serve as a scale for
    plotting any other gel simulation.

    Parameters
    ----------
    bands
        A dictionary of the form {dna_size: migration distance}.
    """

    def __init__(self, bands, name=None, infos=None):
        self.bands = bands
        self.dna_sizes, self.migration_distances = (
            np.array(e) for e in zip(*sorted(bands.items()))
        )
        self._dna_size_to_migration_interpolator = CubicSpline(
            self.dna_sizes, self.migration_distances, bc_type="natural"
        )
        self.infos = infos

    def dna_size_to_migration(self, dna_sizes):
        """Return the migration distances for the given DNA sizes."""
        return self._dna_size_to_migration_interpolator(dna_sizes)

    @property
    def migration_distances_span(self):
        return [
            self.dna_size_to_migration(band)
            for band in (self.dna_sizes.max(), self.dna_sizes.min())
        ]

    def __repr__(self):
        """Represent."""
        return "Ladder(%d-%d)" % (min(self.dna_sizes), max(self.dna_sizes))

    def __str__(self):
        """Represent."""
        return "Ladder(%d-%d)" % (min(self.dna_sizes), max(self.dna_sizes))



# MIT License

# Copyright (c) 2017 Edinburgh Genome Foundry, University of Edinburgh

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Generic class to solve minimal set cover problems.

The class offers methods useful when the coverage is determined by thresholding
scores (see below).

**Technical explanations:**

The problem of enzyme digestion can be generalized and formalized as follows:

A set of parameters Pi
A set of elements ej
A score function score(Pi, ej) => some positive score

The coverage of Pi with threshold T:
C[Pi](T) = [elements e such that score(Pi, e) > T]

The minimal set cover for a threshold T is defined as
M(T) = [Px, Py...] such that [C[Px](T), C[Py](T)...] is minimal set cover of
all elements

Two problems:
A. Find M(T) given T.
B. Find the largest T such that the lenght of M(T) is less than some number N.

Problem A is solved by a standard greedy selection:

.. code::

  greedy_selection(T):
    compute all  C[Pi](T)
    while all elements not covered:
      select a new element.



Problem 2 is solved by bisection on T:

.. code::

  T_minimax = min_Pi(max_ej score(Pi, e_j))
  Tmin, Tmax = 0, T_minimax
  while (Tmax - Tmin) > tolerance:
      Tcenter = (Tmin + Tmax) / 2
      M_center = greedy_selection(Tcenter)
      if length(M_center) > N:
          Tmax = Tcenter
      else:
          Tmin = Tcenter
  return greedy_selection(Tmin)

"""


def maximizing_bisection(f, x_min, x_max, precision=0.001, score=None):
    """Find largest x such that ``score(f(x)) <= 0`` (score(f(x)) increasing).

    Parameters
    ----------
    f
        A function f(x) => anything. The returned value must be either a score,
        or something that can be scored via the ``score`` parameter.
    x_min, x_max
        Range of allowed values for variable x.
    precision
        Precision on x (the algorithm stops when x is less than this far away
        from the optimum).
    score
        Scoring function. Leave to none if function f(x) is already a score.

    Returns
    -------
    x, f(x), score(f(x))
        Everything you may need!
    """
    if score is None:

        def score(value):
            return value

    f_xmin = f(x_min)
    score_xmin = score(f_xmin)
    if score_xmin > 0:
        return None, None, None
    f_xmax = f(x_max)
    score_xmax = score(f_xmax)
    if score_xmax <= 0:
        return x_max, f_xmax, score_xmax

    while (x_max - x_min) > precision:
        center = (x_min + x_max) / 2.0
        f_center = f(center)
        score_center = score(f_center)
        if score_center > 0:
            x_max = center
        else:
            x_min = center
            f_xmin = f_center
            score_xmin = score_center

    return x_min, f_xmin, score_xmin


def minimal_cover(
    elements_set, subsets, max_subsets=None, heuristic="default", selected=(), depth=0
):
    """
    Parameters
    ----------
    elements_set
        The set of all ements to cover.
    subsets
        A list of (name, subset).
    max_subsets
        Maximal number of subsets allowed.
    heuristic
        A function ``((name, subset), selected) => value`` where ``name`` is the
        name of a subset, ``subset`` is what remains of the subset at this stage,
        ``selected`` is a list of already-selected subset names.
    selected
        (Recursion parameter, do not use.) Already-selected elements.
    depth
        (Recursion parameter, do not use.) Depth of the recursion.

    Returns
    -------
        None if no solution was found, else a collection of [(name, subset)...]
        in the order in which the subsets.
    """

    if len(elements_set) == 0:
        return []
    if max_subsets == 0:
        return None

    if depth == 0:
        full_set = set().union(*[subset for name, subset in subsets])
        if full_set != elements_set:
            return None

    subsets = [(n, s) for (n, s) in subsets if len(s)]

    def sorting_heuristic(named_subset):
        name, subset = named_subset
        if heuristic == "default":
            return len(subset)
        else:
            return heuristic(named_subset, selected)

    ordered_subsets = sorted(subsets, key=sorting_heuristic)

    while len(ordered_subsets):
        if max_subsets is not None:
            critical_subset_length = len(elements_set) / max_subsets
            max_len = max(len(s) for name, s in ordered_subsets)
            if max_len < critical_subset_length:
                return None
        name, subset = ordered_subsets.pop()
        new_elements_set = elements_set.difference(subset)
        new_subsets = [
            (name_, sub.difference(subset)) for (name_, sub) in ordered_subsets
        ]
        new_max_subsets = None if (max_subsets is None) else max_subsets - 1
        result = minimal_cover(
            new_elements_set,
            new_subsets,
            heuristic=heuristic,
            selected=list(selected) + [subset],
            max_subsets=new_max_subsets,
            depth=depth + 1,
        )
        if result is not None:
            return result + [name]
        ordered_subsets = [
            subset_
            for (subset_, (new_name, new_subset)) in zip(ordered_subsets, new_subsets)
            if len(new_subset) != 0
        ]
    return None


class SetCoverProblem:
    """Generic class for the resolution of score-based set cover problems.

    Parameters
    ----------
    elements
        A set of all elements to be covered.
    parameters
        A set of parameters (a parameter and a threshold will determine a subset).
    process_logger
        A ProgLog process logger.

    Note
    ----
    The somwhat complicated math problem solved by the algorithms implemented
    in this class is:

    Given set of parameters (Pi), a set of elements (ej), and a score function
    score(Pi, ej) => some positive score, we call "T-coverage of Pi" the set
    ``C[Pi](T) = (elements e such that score(Pi, e) > T)``

    The minimal set cover M(T) is defined as
    M(T) = [Px, Py...] such that [C[Px](T), C[Py](T)...] is minimal subset
    cover of all elements (ej)

    The class allows to solve the two following problems::
    1. Find M(T) given T.
    2. Find the largest T such that the lenght of M(T) is under some number N.
    """

    def __init__(self, elements, parameters, progress_logger=None):
        self.elements = elements
        self.parameters = parameters
        if progress_logger is None:
            progress_logger = lambda **k: None
            progress_logger.iter_bar = lambda **k: k.popitem()[1]
        self.progress_logger = progress_logger
        self._compute_scores()

    @staticmethod
    def _default_heuristic(named_subset, current_selection):
        name, subset = named_subset
        return len(subset)

    def _parameter_element_score(self, parameter, element):
        raise NotImplementedError(
            "Each subset cover problem must overwrite " "this function."
        )

    def _compute_scores(self):
        self.scores = {element: {} for element in self.elements}
        for i, element in enumerate(self.elements):
            self.progress_logger(element_index=i)
            for j, parameter in enumerate(self.parameters):
                score = self._parameter_element_score(parameter, element)
                self.scores[element][parameter] = score

        for parameters_dict in self.scores.values():
            parameters_dict["MAX_SCORE"] = max(
                (score, parameter) for (parameter, score) in parameters_dict.items()
            )
        ((min_max_score, parameter), element) = min(
            (parameters_dict["MAX_SCORE"], element)
            for element, parameters_dict in self.scores.items()
        )
        self.bottleneck = dict(
            score=min_max_score, parameter=parameter, element=element
        )

    def _compute_coverages(self, threshold):
        return {
            parameter: set(
                element
                for element in self.elements
                if self.scores[element][parameter] >= threshold
            )
            for parameter in self.parameters
        }

    def _select_parameters(
        self,
        threshold=None,
        max_set_size=None,
        covering_algorithm="greedy",
        heuristic="default",
        threshold_tolerance=0.001,
        bisection=True,
    ):
        """Select a subset of parameters which collectively cover all elements.

        This lets you either find the highest-score cover of less than N
        elements, or the best free-sized set of parameters covering elements
        with the highest minimal score.

        Parameters
        ----------
        threshold
            If provided and max_set_size is not provided, the method will return
            a free-sized set of parameters covering all elements. If max_set_size
            is provided, the subset will not exceed max_set_size in size. This
            can result in unsolvable.
        max_set_size
            When provided instead of threshold, the method will find the highest
            threshold which allows solutions in max_set_size or less.
        covering_algorithm
            Either 'greedy' for fast approximate solving or 'full' for full
            solving.
        heuristic
            Scoring function to prioritize some parameters over others.
        threshold_tolerance
            When max_set_size is provided and not threshold, the returned
            solution will have the optimal threshold, plus or minus this
            tolerance.
        bisection
            For internal recursion. Do not modify.

        Returns
        -------
        min_score, [param1, param2, param3...]
            Where min_score is a value lower than all the score coverages, and
            the list indicates the parameters in the order in which they have
            been selected.
        """

        if heuristic == "default":
            heuristic = self._default_heuristic
        if threshold is None:
            threshold = self.bottleneck["score"]

        if max_set_size is not None:
            if not bisection:
                return threshold, minimal_cover(
                    elements_set=self.elements,
                    subsets=self._compute_coverages(threshold).items(),
                    max_subsets=max_set_size,
                    heuristic=heuristic,
                )

            def select(thr):
                _, selection = self._select_parameters(
                    threshold=thr,
                    covering_algorithm=covering_algorithm,
                    heuristic=heuristic,
                    max_set_size=(
                        None if covering_algorithm == "greedy" else max_set_size
                    ),
                    bisection=False,
                )
                return selection

            def score(selection):
                if selection is None:
                    return 1e10
                else:
                    return len(selection) - max_set_size

            threshold, selection, _ = maximizing_bisection(
                f=select,
                x_min=0,
                x_max=threshold,
                score=score,
                precision=threshold_tolerance,
            )
            return threshold, selection
        else:
            if covering_algorithm == "greedy":
                max_subsets = None
            else:
                max_subsets = max_set_size
            return threshold, minimal_cover(
                elements_set=self.elements,
                subsets=self._compute_coverages(threshold).items(),
                max_subsets=max_subsets,
                heuristic=heuristic,
            )

    def min_parameter_score(self, parameter):
        """Return the smallest score on any element for the given parameter."""
        return min([self.scores[e][parameter] for e in self.elements])



from collections import OrderedDict
import itertools

import numpy as np

from .DigestionProblem import DigestionProblem
from ..tools import max_min_distance

try:
    import matplotlib.pyplot as plt

    PLOTS_AVAILABLE = True
except ImportError:
    PLOTS_AVAILABLE = False


class SeparatingDigestionsProblem(DigestionProblem):
    """Problem: find best digestion(s) to identify constructs.

    Provided a set of constructs (possibly from a combinatorial assembly),
    find a list of digestions such that all pair of constructs have different
    migration patterns for at least one digestion of the list.

    Parameters
    ----------
    enzymes
        List of the names of the enzymes to consider, e.g. ``['EcoRI', 'XbaI']``.
    ladder
        A Ladder object representing the ladder used for migrations.
    sequences
        An (ordered) dictionary of the form {sequence_name: sequence} where the
        sequence is an ATGC string.
    topology
    max_enzymes_per_digestion
        Maximum number of enzymes that can go in a single digestion.
        Experimentally the best is 1, but you can try 2, or 3 in desperate
        situations. Not sure if more enzymes will work.
    min_discrepancy
    relative_migration_precision
        Variance of the bands measured during the migration, given as a
        proportion of the total migration span (difference between the migration
        of the ladder's smallest and largest bands).
    """

    def __init__(
        self,
        enzymes,
        ladder,
        sequences=None,
        categories=None,
        topology="default_to_linear",
        max_enzymes_per_digestion=1,
        min_discrepancy="auto",
        relative_migration_precision=0.1,
    ):
        """Initialize."""
        if isinstance(sequences, (list, tuple)):
            sequences = OrderedDict([(r.id, r) for r in sequences])

        if categories is None:
            categories = OrderedDict(
                [
                    (seq_name, {seq_name: sequence})
                    for seq_name, sequence in sequences.items()
                ]
            )
        else:
            sequences = OrderedDict(
                [
                    (seq_name, sequence)
                    for category, seqs in categories.items()
                    for seq_name, sequence in seqs.items()
                ]
            )
        self.categories = categories

        DigestionProblem.__init__(
            self,
            sequences=sequences,
            enzymes=enzymes,
            ladder=ladder,
            topology=topology,
            max_enzymes_per_digestion=max_enzymes_per_digestion,
            relative_migration_precision=relative_migration_precision,
        )

    def _compute_elements(self):
        category_pairs = itertools.combinations(self.categories.values(), 2)
        return set(
            (seq1, seq2)
            for category1, category2 in category_pairs
            for seq1 in category1
            for seq2 in category2
        )

    def _parameter_element_score(self, digestion, sequences_pair):
        """See max_patterns_difference."""
        sequence1, sequence2 = sequences_pair
        digestion1, digestion2 = [
            self.sequences_digestions[s][digestion] for s in (sequence1, sequence2)
        ]
        # If similar pair already computed, return the previous result.
        if digestion1["same_as"] == digestion2["same_as"]:
            if digestion1["same_as"] in self.scores[sequences_pair]:
                return self.scores[sequences_pair][digestion1["same_as"]]

        mini, maxi = self.migration_min, self.migration_max
        margin = self.relative_migration_precision * self.migration_span / 2.0
        distance = max_min_distance(
            digestion1["migration"],
            digestion2["migration"],
            zone=[mini + margin, maxi - margin],
        )
        return 1.0 * distance / self.migration_span

    @staticmethod
    def _score_to_color(score, maxi=0.1):
        """Transform a similarity score to a green/red color.

        Parameters
        ----------
        score
            Value between 0 (perfect similarity, green) and 1 (red).
        maxi
            Value of the score above which everything appears completely red.
            Below this value the color goes progressively from red to green in 0.
        """
        return (
            max(0, min(1, score / maxi)),
            min(1, max(0, 1 - score / maxi)),
            0,
            0.5,
        )

    def plot_distances_map(self, digestions, ax=None, target_file=None):
        """Plot how well the digestions separate each construct pair.

        Parameters
        ----------
        digestions
            A list of digestions, eg ``[('EcoRV'), ('XbaI', 'MfeI')]``.
        ax
            A matplotlib ax on which to plot, if none is provided, one is created
            and returned at the end.
        target_file
            The name of the (PNG, SVG, JPEG...) file in which to write the plot.

        Returns
        -------
        axes
            The axes of the generated figure (if a target file is written to,
            the figure is closed and None is returned instead).
        """

        if not PLOTS_AVAILABLE:
            raise ImportError("Plots require Matplotlib/Bandwagon installed.")
        grid = np.zeros(2 * (len(self.sequences),))
        for i, s1 in enumerate(self.sequences):
            for j, s2 in enumerate(self.sequences):
                if i >= j:
                    grid[i, j] = np.nan
                else:
                    scores = [
                        self._parameter_element_score(digestion, (s1, s2))
                        for digestion in digestions
                    ]
                    grid[i, j] = max(scores)
        if ax is None:
            _, ax = plt.subplots(1, figsize=2 * (0.8 * len(grid),))
        ax.imshow(
            grid[:, ::-1],
            interpolation="nearest",
            cmap="OrRd_r",
            vmin=0,
            vmax=self.relative_migration_precision,
        )
        for i in range(len(grid)):
            for j in range(len(grid)):
                if i > j:
                    ax.text(
                        len(self.sequences) - i - 1,
                        j,
                        "%d%%" % (100 * grid[j, i]),
                        fontdict=dict(color="black", weight="bold", size=14),
                        horizontalalignment="center",
                        verticalalignment="center",
                    )

        ax.set_yticks(range(len(grid)))
        ax.set_yticklabels(list(self.sequences), size=14, fontdict={"weight": "bold"})
        ax.set_xticks(range(len(grid)))
        ax.xaxis.set_ticks_position("top")
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.set_xticklabels(
            [" " + s for s in list(self.sequences)[::-1]],
            rotation=90,
            size=14,
            fontdict={"weight": "bold"},
        )
        ax.set_xlim(-0.5, len(self.sequences) - 1.5)
        ax.set_ylim(len(self.sequences) - 1.5, -0.5)

        if target_file is not None:
            ax.figure.savefig(target_file, bbox_inches="tight")
            plt.close(ax.figure)
        return ax



from .DigestionProblem import DigestionProblem
import numpy as np


class IdealDigestionsProblem(DigestionProblem):
    """Find ideal digestion(s) to validate constructs.

    Other digestion problems subclass this problem and implement a computation
    of coverages which depends on the problem.

    Parameters
    ----------
    sequences
        An (ordered) dictionary of the form {sequence_name: sequence} where the
        sequence is an ATGC string.
    enzymes
        List of the names of the enzymes to consider, e.g. ``['EcoRI', 'XbaI']``.
    ladder
        A Ladder object representing the ladder used for migrations.
    linear
        True for linear sequences, False for circular sequences.
    max_enzymes_per_digestion
        Maximal number of enzymes that can go in a single digestion.
        Experimentally the best is 1, but you can try 2, or 3 in desperate
        situations. Not sure if more enzymes will work.
    relative_migration_error
        Variance of the bands measured during the migration, given as a
        proportion of the total migration span (difference between the migration
        of the ladder's smallest and largest bands).
    """

    def __init__(
        self,
        enzymes,
        ladder,
        sequences,
        min_bands=3,
        max_bands=7,
        border_tolerance=0.1,
        topology="default_to_linear",
        max_enzymes_per_digestion=1,
        relative_migration_precision=0.1,
    ):
        """Initialize."""
        self.min_bands = min_bands
        self.max_bands = max_bands
        self.border_tolerance = border_tolerance

        DigestionProblem.__init__(
            self,
            sequences=sequences,
            enzymes=enzymes,
            ladder=ladder,
            topology=topology,
            max_enzymes_per_digestion=max_enzymes_per_digestion,
            relative_migration_precision=relative_migration_precision,
        )

    def _parameter_element_score(self, digestion, sequence):
        """Compute the sequence's ``.migration_score`` for each digestion."""
        digestion = self.sequences_digestions[sequence][digestion]
        if digestion["same_as"] in self.scores[sequence]:
            return self.scores[sequence][digestion["same_as"]]
        migration = digestion["migration"]
        return self.migration_score(migration)

    def migration_score(self, band_migrations):
        """Score the well-numbering and well-separation of all bands.

        If some bands are too high or too low, or the number of bands is out
        of bounds, return 0. Else, return the minimal distance between two
        consecutive bands.
        """
        if not self.min_bands <= len(band_migrations) <= self.max_bands:
            return 0

        t = self.border_tolerance
        mini, maxi = (1 + t) * self.migration_min, (1 - t) * self.migration_max
        if not (mini <= min(band_migrations) <= max(band_migrations) < maxi):
            return 0

        min_gap = np.diff(sorted(band_migrations)).min()
        return 1.0 * min_gap / self.migration_span

    def _compute_elements(self):
        """Return the list of digestions to serve as elements."""
        return set(self.sequences.keys())



class CloneValidation:
    """Report from comparing a clone's pband patterns with expectations.

    Instances of this class are produced by validating a ``Clone`` object.

    Parameters
    ----------
    clone
      A Clone object.
    expected
        A dictionary ``{digestion: expected_pattern}`` where ``digestion`` is
        of the form ``('EcoRI', 'BamHI')`` and ``expected_pattern`` is a
        BandsPattern.
    discrepancies
      A dictionary ``{digestion: float}`` where the float indicates the
      observed discrepancy between observed and predicted for this digestion.
    """

    def __init__(self, clone, expected, discrepancies):
        """Initialize."""
        self.clone = clone
        self.expected = expected
        self.discrepancies = discrepancies

    @property
    def passes(self):
        """True if the clone is valid.

        (This means that all patterns are similar enough to expectations.)
        """
        return self.max_discrepancy < 1.0

    @property
    def max_discrepancy(self):
        """Return the largest discrepancy in all patterns."""
        return max(self.discrepancies.values())

    def color(self, discrepancy="max"):
        """Return a color depending on the discrepancy level (in 0-1)."""
        if discrepancy == "max":
            discrepancy = self.max_discrepancy
        factor = min(1.0, discrepancy**10)
        return ((2 + factor) / 3.0, (3 - factor) / 3.0, 2 / 3.0)

    def to_bandwagon_bandpattern(
        self, digestion, label="auto", per_digestion_discrepancy=True
    ):
        """Return a (plottable) bandwagon pattern for the selected digestion.

        The pattern has a color indicating the match similarity, as well as
        a corner note indicating the gap to similarity.

        Parameters
        ----------
        digestion
            A set of enzymes such as ``('EcoRI',)``.
        label
            A label for the pattern, for instance the name of a microplate well.
        per_digestion_discrepancy
            If true, each pattern for each clone will have an indication of
            how close it is from the expected pattern. If False, and the
            validation involves several digestions, the biggest discrepancy
            among all patterns is shown.
        """
        if digestion not in self.clone.digestions:
            return None
        if per_digestion_discrepancy:
            discrepancy = self.discrepancies[digestion]
        else:
            discrepancy = self.max_discrepancy
        obs = self.clone.digestions[digestion]
        pattern = obs.to_bandwagon_bandpattern(
            background_color=self.color(discrepancy=discrepancy), label=label
        )
        pattern.corner_note += " Gap: %d" % (100 * discrepancy)
        return pattern



import time
import tqdm
import pandas
import requests

if __name__ == "__main__":
    website_url = "http://rebase.neb.com"
    methylations = ["Dam", "Dcm", "CpG", "EcoBI", "EcoKI"]
    columns = ["enzyme", "req_seq", "suppliers"] + methylations

    print("Getting enzyme list and suppliers...")

    html = requests.get(website_url + "/cgi-bin/azlist?re2+cy").text
    tables = pandas.read_html(html, header=0)
    table = [t for t in tables if (t.values.shape[1] == 4)][0]
    suppliers_dict = {
        enzyme: len(suppliers)
        for enzyme, suppliers in zip(table.Enzymes, table.Suppliers)
    }

    print("Getting enzymes methylation data...")

    def get_enzyme_data(enzyme_name):
        url = website_url + "/cgi-bin/damlist?e" + enzyme_name
        html = requests.get(url).text
        table = [t for t in pandas.read_html(html) if t.values.shape == (3, 6)][0]
        rec_seq = table.values[0, 0].replace(enzyme_name, "")
        suppliers = suppliers_dict[enzyme_name]
        return dict(
            [("enzyme", enzyme_name), ("req_seq", rec_seq), ("suppliers", suppliers)]
            + list(zip(methylations, table.values[2, 1:6]))
        )

    data = []
    errored = []
    for enzyme_name in tqdm.tqdm(list(suppliers_dict)):
        time.sleep(0.5)
        try:
            data.append(get_enzyme_data(enzyme_name))
        except Exception as e:
            errored.append(enzyme_name)
    if len(errored):
        print("%d errors:" % len(errored), " ".join(errored))

    print("Generating the spreadsheet...")

    dataframe = pandas.DataFrame(data, columns=columns)
    dataframe.to_csv("./enzymes_infos.csv", index=False)

    print("Done !")



"""
Complex validation test.
"""

import matplotlib

matplotlib.use("Agg")

import os
from bandwitch import ClonesObservations


def test_complex_validation(tmpdir):
    data_dir = os.path.join("tests", "test_data", "complex_validation_data")
    records_path = os.path.join(data_dir, "constructs_sequences.zip")
    constructs_map_path = os.path.join(data_dir, "constructs_map.xls")
    digestions_map_path = os.path.join(data_dir, "digestions_map.xls")
    aati_zip_path = os.path.join(data_dir, "digestion_results.zip")
    clones = ClonesObservations.from_files(
        records_path=records_path,
        constructs_map_path=constructs_map_path,
        aati_zip_path=aati_zip_path,
        digestions_map_path=digestions_map_path,
    )
    validations = clones.validate_all_clones(relative_tolerance=0.03)
    clones.plot_validations_plate_map(validations)
    clones.plot_all_validations_patterns(validations)
    partial_digest_analysis = clones.partial_digests_analysis()
    clones.plot_partial_digests_analysis(partial_digest_analysis)
    summary_table = clones.validations_summary_table(
        validations, target=os.path.join(str(tmpdir), "table.csv")
    )


# Chosen digestions, for reference:
# C00	EcoRI,PvuI
# C01	EcoRI,PvuI
# C02	EcoRI,PvuI
# C03	EcoRI,PvuI
# C04	EcoRI,PvuI
# C05	EcoRI,PvuI
# C06	AseI,EcoRI
# C07	AseI,EcoRI
# C08	AseI,EcoRI



from collections import OrderedDict
from bandwagon.ladders import ladder_from_aati_fa_calibration_table
from bandwagon import BandsPattern
from .band_patterns_discrepancy import band_patterns_discrepancy

try:
    from plateo.parsers import plate_from_aati_fragment_analyzer_zip

    PLATEO_AVAILABLE = True
except ImportError:
    PLATEO_AVAILABLE = False


class BandsObservation:
    """One observation of a bands pattern.

    Parameters
    ----------
    name : str
        Name of the observation (used for the top label in plots).
    bands : BandPattern
        A BandPattern object.
    ladder : BandPattern
        A BandPattern object representing a ladder.
    migration_image
        Optional RGB array (HxWx3) representing the gel "image" (which will be
        displayed on the side of the plot).
    """

    def __init__(self, name, bands, ladder, migration_image=None):
        """Initialize."""
        self.name = name
        self.bands = sorted(bands)
        self.ladder = ladder
        self.migration_image = migration_image

    @staticmethod
    def from_aati_fa_archive(
        archive_path,
        min_rfu_size_ratio=0.3,
        ignore_bands_under=None,
        direction="column",
    ):
        """Return a dictionary of all band observations in AATI output files.

        Parameters
        ----------
        archive_path : str
            A path to a ZIP file containing all files output by the AATI fragment
            analyzer.
        min_rfu_size_ratio : float
            Cut-off ratio to filter out bands whose intensity is below some
            threshold. The higher the value, the more bands will be filtered out.

        Returns
        -------
        dict
            A dictionary ``{'A1': BandsObservation(), 'A2': ...}`` containing the
            measured pattern information for a whole 96-well microplate.
        """
        if not PLATEO_AVAILABLE:
            raise ImportError("Plateo must be installed to parse AATI zips.")

        ignore_bands_under = ignore_bands_under or 0

        plate = plate_from_aati_fragment_analyzer_zip(archive_path)
        ladder_data = plate.data["ladder"]
        ladder = ladder_from_aati_fa_calibration_table(dataframe=ladder_data)
        # get the set of constructs sorted by order of columnwise appearance:

        def band_is_strong_enough(band):
            """Return True if the band's intensity is above the set level."""
            return 1.0 * band["RFU"] / band["Size (bp)"] > min_rfu_size_ratio

        ordered_list = []
        for well in plate.iter_wells(direction=direction):
            bands_list = []
            for band in well.data["bands"].values():
                if band_is_strong_enough(band) and (
                    band["Size (bp)"] > ignore_bands_under
                ):
                    bands_list += [band["Size (bp)"]]

            bands_observation = BandsObservation(
                name=well.name,
                ladder=ladder,
                bands=bands_list,
                migration_image=well.data["migration_image"],
            )

            entry = (well.name, bands_observation)
            ordered_list += [entry]

        ordered_dict = OrderedDict(ordered_list)

        return ordered_dict

    def patterns_discrepancy(
        self,
        other_bands,
        relative_tolerance=0.1,
        min_band_cutoff=None,
        max_band_cutoff=None,
    ):
        """Return the maximal discrepancy between two band patterns.

        The discrepancy is defined as the largest distance between a band in
        one pattern and the closest band in the other pattern.

        Parameters
        ----------
        other_bands : list of int
            A list of bands (integers) to be compared with the current bands.
        relative_tolerance : float
            Tolerance, as a ratio of the full ladder span. If =0.1, then the
            discrepancy will have a value of 1 when a band's nearest
            correspondent in the other pattern is more than 10% of the ladder
            span apart.
        min_band_cutoff : int, optional
            Discrepancies involving at least one band below this minimal band
            size will be ignored. By default, it will be set to the smallest
            band size in the ladder.
        max_band_cutoff : int, optional
            Discrepancies involving at least one band above this minimum band
            size will be ignored. By default, it will be set to the largest
            band size in the ladder.
        """
        ladder_min, ladder_max = self.ladder.dna_size_span
        if min_band_cutoff is None:
            min_band_cutoff = ladder_min
        if max_band_cutoff is None:
            max_band_cutoff = ladder_max
        return band_patterns_discrepancy(
            other_bands,
            self.bands,
            ladder=self.ladder,
            relative_tolerance=relative_tolerance,
            zone=[min_band_cutoff, max_band_cutoff],
            reference_and_gel=True,
        )

    def to_bandwagon_bandpattern(self, background_color=None, label="auto"):
        """Return a pattern version for the plotting library BandWagon.

        If label is left to 'auto', it will be the pattern's name.
        """
        if label == "auto":
            label = self.name
        return BandsPattern(
            self.bands,
            corner_note="Total: %d bp." % sum(self.bands),
            ladder=self.ladder,
            label=label,
            gel_image=self.migration_image,
            background_color=background_color,
        )



"""Implements the DigestionProblem class."""

from collections import OrderedDict
from copy import deepcopy

import numpy as np

from ..tools import (
    updated_dict,
    sequence_to_biopython_record,
    record_is_linear,
    set_record_topology,
)
from .SetCoverProblem import SetCoverProblem
from ..bands_predictions import (
    predict_sequence_digestions,
    compute_sequence_digestions_migrations,
)

try:
    import bandwagon
    import matplotlib.pyplot as plt

    PLOTS_AVAILABLE = True
except ImportError:
    PLOTS_AVAILABLE = False


class DigestionProblem(SetCoverProblem):
    """General class for solving digestions problems.

    By digestion problem we mean *enzymatic* digestion problems. For other
    kinds of digestion problems, see a proctologist.

    Other digestion problems subclass this problem for particular cases such
    as construct validation and identification.

    Parameters
    ----------
    enzymes
        List of the names of the enzymes to consider, e.g. ``['EcoRI', 'XbaI']``.
    ladder
          A Ladder object representing the ladder used for migrations.
    sequences
          An (ordered) dictionary of the form {sequence_name: sequence} where the
          sequence is an ATGC string.
    topology
        The records opology. Either "circular", "linear", "default_to_circular"
        (will default to circular if ``record.annotations['topology']`` is not
        already set) or "default_to_linear".
    max_enzymes_per_digestion
          Maximum number of enzymes that can go in a single digestion.
          Experimentally the best is 1, but you can try 2, or 3 in desperate
          situations. Not sure if more enzymes will work.
    relative_migration_precision
        Variance of the bands measured during the migration, given as a
        proportion of the total migration span (difference between the migration
        of the ladder's smallest and largest bands).
    progress_logger
    """

    def __init__(
        self,
        enzymes,
        ladder,
        sequences,
        topology="default_to_linear",
        max_enzymes_per_digestion=1,
        relative_migration_precision=0.1,
        progress_logger=None,
    ):
        """Initialize."""
        if hasattr(sequences, "items"):
            if hasattr(list(sequences.values())[0], "seq"):
                self.records = {
                    name: deepcopy(record) for name, record in sequences.items()
                }
                self.sequences = {
                    name: str(record.seq) for name, record in sequences.items()
                }
            else:
                self.sequences = sequences
                self.records = {
                    name: sequence_to_biopython_record(seq)
                    for name, seq in sequences.items()
                }
        else:
            self.records = OrderedDict([(r.id, deepcopy(r)) for r in sequences])
            self.sequences = OrderedDict([(r.id, str(r.seq)) for r in sequences])
        for record in self.records.values():
            set_record_topology(record, topology)
        self.relative_migration_precision = relative_migration_precision
        self.max_enzymes_per_digestion = max_enzymes_per_digestion

        self.sequences_names = list(self.sequences.keys())
        self.progress_logger = progress_logger
        self.enzymes = enzymes

        self.sequences_digestions = {
            name: predict_sequence_digestions(
                sequence=sequence,
                enzymes=self.enzymes,
                linear=record_is_linear(self.records[name]),
                max_enzymes_per_digestion=self.max_enzymes_per_digestion,
            )
            for name, sequence in self.sequences.items()
        }
        digestions = self.sequences_digestions[self.sequences_names[0]]
        self.digestions = list(digestions.keys())
        self.change_ladder(ladder)

    def change_ladder(self, ladder):
        self.ladder = ladder
        mini, maxi = self.ladder.migration_distances_span
        self.migration_min, self.migration_max = mini, maxi
        self.migration_span = maxi - mini
        compute_sequence_digestions_migrations(self.sequences_digestions, self.ladder)
        SetCoverProblem.__init__(
            self,
            elements=self._compute_elements(),
            parameters=self.digestions,
            progress_logger=self.progress_logger,
        )

    @staticmethod
    def _default_heuristic(named_subset, selected):
        """Select for coverage first, digestion complexity second.

        When two digestions cover the same number of elements, the digestion
        with the fewer enzymes is preferred.
        """
        enzymes, subset = named_subset
        return len(subset) - 0.5 * len(enzymes)

    def bands_to_migration_pattern(self, bands_sizes):
        """Return the distance migrations from several band sizes."""
        return self.ladder.dna_size_to_migration(np.array(bands_sizes))

    def plot_digestions(
        self,
        digestions,
        axes=None,
        bands_props=None,
        patterns_props=None,
        patternset_props=None,
        target_file=None,
        close_figure=False,
    ):
        """Plot the patterns for each sequence, for each digestion in the list.

        Requires BandWagon.

        Parameters
        ----------
        digestions
              A list of digestions, e.g. [('EcoRI',), ('BamHI', 'XbaI'), ...].
        axes
            Axes on which to plot the plot. There should be as many axes in the
            list as there are digestions provided. Ideally the axes would be
            vertically stacked. If not provided, new axes are created and
            returned in the end.
        bands_props, patterns_props, patternset_props
            Graphical properties (colors, labels, etc.) of band patterns, see
            code and BandWagon for more details.
        target_file
             The name of the (PNG, SVG, JPEG...) file in which to write the plot.
        close_figure

        Returns
        -------
        axes
          The axes of the generated figure (if a target file is written to,
          the figure is closed and None is returned instead).
        """
        if not PLOTS_AVAILABLE:
            raise ImportError("Plotting requires Bandwagon installed")

        bands_props = updated_dict({"size": 10, "rotation": 50}, bands_props)
        patternset_props = updated_dict({"ladder_ticks": 5}, patternset_props)
        if isinstance(self.ladder.bands, (list, tuple)):
            ladder = self.ladder
        else:
            ladder = bandwagon.custom_ladder(None, self.ladder.bands)
        ndig, nseq = len(digestions), len(self.sequences)
        if axes is None:
            fig, axes = plt.subplots(ndig, 1, figsize=(0.5 * nseq, 3 * ndig))
        if ndig == 1:
            axes = [axes]
        for ax, digestion in zip(axes, digestions):
            bandwagon.BandsPatternsSet(
                patterns=[ladder if ax == axes[0] else ladder.modified(label=None)]
                + [
                    bandwagon.BandsPattern(
                        bands=self.sequences_digestions[seq_name][digestion]["bands"],
                        label=seq_name if (ax == axes[0]) else None,
                        ladder=ladder,
                        global_bands_props=bands_props,
                        topology=self.records[seq_name].annotations["topology"],
                    )
                    for seq_name in self.sequences
                ],
                ladder=ladder,
                label=", ".join([" + ".join([e for e in d]) for d in [digestion]]),
                global_patterns_props=patterns_props,
                **patternset_props
            ).plot(ax)
        if target_file is not None:
            axes[0].figure.savefig(target_file, bbox_inches="tight")
        if close_figure:
            plt.close(axes[0].figure)
        return axes

    def select_digestions(
        self,
        minimal_score=None,
        max_digestions=None,
        search="greedy",
        min_score_precision=0.001,
    ):
        """Select one or more digestions which collectively solve the problem.

        This lets you either find the highest-score solution of less than N
        digestions, or the best free-sized set of digestions solving the
        problem with the highest minimal score.

        Returns
        -------
        min_score, [digestion1, digestion2, ...]
            Where min_score is the lowest score that an element (sequence or
            sequence pair) gets from the coverage, and the list indicates the
            digestions in the order in which they have been selected. This list
            will be "None" for unsolvable problems.

        Parameters
        -----------
        minimal_score
            If provided and max_digestions is not provided, the method will
            return a free-sized set of parameters covering all elements.
            If max_digestions is provided, the subset will not exceed
            max_digestions in size. This can result in unsolvable problems.
        max_digestions
            When provided instead of minimal_score, the method will find the
            highest minimal_score which allows solutions in max_digestions or
            less.
        search
            Either 'greedy' for fast approximate solving or 'full' for full
            solving.
        min_score_precision
            When max_digestions is provided and not minimal_score, the returned
            solution will have the optimal minimal_score, plus or minus this
            tolerance.
        """
        return self._select_parameters(
            threshold=minimal_score,
            covering_algorithm=search,
            max_set_size=max_digestions,
            heuristic="default",
            threshold_tolerance=min_score_precision,
            bisection=True,
        )



"""Loads REBASE enzymes infos (mehtylation sensitivity and providers)"""

__all__ = ["enzymes_infos"]

import os.path as osp

csv_path = osp.join(osp.dirname(osp.realpath(__file__)), "enzymes_infos.csv")

with open(csv_path, "r") as f:
    _lines = f.read().split("\n")
    _fields = _lines[0].split(";")
    _replacements = dict(
        [("N/A", False), ("+", True), ("-", True)] + [(str(i), i) for i in range(50)]
    )
    enzymes_infos = {
        _line.split(";")[0]: dict(
            zip(_fields, [_replacements.get(e, e) for e in _line.split(";")])
        )
        for _line in _lines[1:]
    }



import os
from .Ladder import Ladder

this_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)))
ladders_path = os.path.join(this_dir, "data")
infos_file = "ladders_infos.csv"
with open(os.path.join(ladders_path, infos_file), "r") as f:
    infos = dict(line.split(";") for line in f.read().split("\n")[1:])


class LadderDict(dict):
    pass


LADDERS = LadderDict()
for filename in os.listdir(ladders_path):
    if filename == infos_file:
        continue
    ladder_name = filename.split(".")[0]
    with open(os.path.join(ladders_path, filename), "r") as f:
        ladder = Ladder(
            infos=infos[ladder_name],
            name=ladder_name,
            bands=dict(
                tuple(int(e) for e in line.split(";"))
                for line in f.read().split("\n")[1:]
            ),
        )
    LADDERS[ladder_name] = LADDERS.__dict__["ladder_" + ladder_name] = ladder



"""Module for digestion pattern prediction."""

from Bio import Restriction
from Bio.Seq import Seq
from collections import OrderedDict
import numpy as np

from .tools import all_subsets


def _compute_bands_from_cuts(cuts, sequence_length, linear=True):
    """Compute the size of the obtained bands from the position of cuts.

    Returns a list of band sizes.

    Parameters
    ----------
    cuts
        Location of the different cuts on the plasmid.
    sequence_length
        Length of the DNA molecule.
    linear
        True for a linear DNA molecule, False for a circular DNA molecule.
    """
    cuts = [0] + sorted(list(set(cuts))) + [sequence_length]
    bands = [cut2 - cut1 for cut1, cut2 in zip(cuts, cuts[1:])]
    if not linear and len(bands) > 1:
        bands[0] += bands.pop()
    return bands


def find_cuts(sequence, enzymes, linear=True):
    batch = Restriction.RestrictionBatch(enzymes)
    return [
        cut for cuts in batch.search(sequence, linear=linear).values() for cut in cuts
    ]


def predict_digestion_bands(sequence, enzymes, linear=True, partial_cutters=()):
    """Return the band sizes from digestion by all enzymes at once.

    Returns a list of bands sizes sorted from smallest to largest.

    Parameters
    ----------
    sequence
        Sequence to be digested. Either a string ("ATGC...") or
        a Biopython `Seq`.
    enzymes
        list of all enzymes placed at the same time in the digestion
        mix e.g. `["EcoRI", "BamHI"]`.
    linear
        True if the DNA fragment is linearized, False if it is circular.
    partial_cutters
        List of enzymes that are tired or inhibited would randomly miss
        some sites, thus creating extra bands in the final digest.
    """
    if not isinstance(sequence, Seq):
        sequence = Seq(sequence)
    cuts = find_cuts(sequence, enzymes, linear=linear)
    if len(partial_cutters) == 0:
        bands = _compute_bands_from_cuts(cuts, len(sequence), linear=linear)
    else:
        propartial_cuts = find_cuts(sequence, partial_cutters, linear=linear)
        bands = set(
            band
            for partial_cuts in all_subsets(propartial_cuts)
            for band in _compute_bands_from_cuts(
                cuts + list(partial_cuts), len(sequence), linear
            )
        )
    return sorted(bands)


def compute_sequence_digestions_migrations(sequences_digestions, ladder):
    """Add a 'migration' field to the data in sequences_digestions.

    ``sequences_digestions`` is a dict ``{seq: digestions_data}``
    where ``digestions_data`` is a result of ``predict_sequence_digestions``,
    ``ladder`` is a Ladder object.
    """
    all_bands = set()
    for seq, digestions_data in sequences_digestions.items():
        for digestion, data in digestions_data.items():
            all_bands.update(data["bands"])
    all_bands = sorted(all_bands)
    all_migrations = ladder.dna_size_to_migration(list(all_bands))
    band_to_migration = {
        band: migration for band, migration in zip(all_bands, all_migrations)
    }
    for seq, digestions_data in sequences_digestions.items():
        for digestion, data in digestions_data.items():
            data["migration"] = np.array([band_to_migration[b] for b in data["bands"]])


def predict_sequence_digestions(
    sequence, enzymes, linear=True, max_enzymes_per_digestion=1
):
    """Return a dict giving bands sizes pattern for all possible digestions.

    The digestions, double-digestions, etc. are listed and for each the
    sequence band sizes are computed.

    The result if of the form ``{digestion: {'cuts': [], 'bands': []}}``
    Where ``digestion`` is a tuple of enzyme names e.g. ``('EcoRI', 'XbaI')``,
    'cuts' is a list of cuts locations, 'bands' is a list of bands sizes.

    Parameters
    ----------
    sequence
        The sequence to be digested.
    enzymes
        List of all enzymes to be considered.
    max_enzymes_per_digestion
        Maximum number of enzymes allowed in one digestion.
    bands_to_migration
        Function associating a migration distance to a band size. If provided,
        each digestion will have a ``'migration'`` field (list of migration
        distances) in addition to 'cuts' and 'bands'.
    """
    restriction_batch = Restriction.RestrictionBatch(enzymes)
    cuts_dict = restriction_batch.search(Seq(sequence))

    def get_cuts(enzyme_name):
        return {"cuts": cuts_dict[Restriction.__dict__[enzyme_name]]}

    def _merge_digestions(digestion1, digestion2, sequence_length, linear):
        """Merge and sort the cuts from two different digestions."""
        all_cuts = sorted(list(set(digestion1["cuts"] + digestion2["cuts"])))
        return {
            "cuts": all_cuts,
            "bands": _compute_bands_from_cuts(
                cuts=all_cuts, sequence_length=sequence_length, linear=linear
            ),
        }

    empty_digestion = ((), {"cuts": [], "bands": [len(sequence)]})
    digestions_dict = OrderedDict([empty_digestion])
    for n_enzymes in range(max_enzymes_per_digestion):
        sub_enzymes = [
            enzs for enzs in digestions_dict.keys() if len(enzs) == n_enzymes
        ]
        for enzyme in enzymes:
            sub_sub_enzymes = [enzs for enzs in sub_enzymes if enzyme not in enzs]
            for enzs in sub_sub_enzymes:
                digestion = tuple(sorted(enzs + (enzyme,)))
                if digestion not in digestions_dict:
                    no_enzyme_band = len(get_cuts(enzyme)["cuts"]) == 0
                    no_enzs_band = len(digestions_dict[enzs]["cuts"]) == 0
                    one_no_bands = no_enzs_band or no_enzyme_band
                    if ((enzyme,) in digestions_dict) and one_no_bands:
                        if no_enzyme_band:
                            digestions_dict[digestion] = digestions_dict[enzs]
                            digestions_dict[digestion]["same_as"] = enzs
                        elif no_enzs_band:
                            dig = (enzyme,)
                            digestions_dict[digestion] = digestions_dict[dig]
                            digestions_dict[digestion]["same_as"] = dig
                    else:
                        digestions_dict[digestion] = _merge_digestions(
                            digestion1=get_cuts(enzyme),
                            digestion2=digestions_dict[enzs],
                            sequence_length=len(sequence),
                            linear=linear,
                        )
                        digestions_dict[digestion]["same_as"] = digestion
    digestions_dict.pop(())

    # Reordering the dictionnary makes the computation of scores faster
    # using 'same_as' to avoid recomputing scores involving similar patterns
    digestions_dict = OrderedDict(
        sorted(
            digestions_dict.items(),
            key=lambda item: (len(item[0]), len(item[1]["cuts"])),
        )
    )
    return digestions_dict



"""Collection of useful methods and solvers for BandWitch."""

import itertools
import os

import numpy as np
from Bio.Seq import Seq
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

try:
    # Biopython <1.78
    from Bio.Alphabet import DNAAlphabet

    has_dna_alphabet = True
except ImportError:
    # Biopython >=1.78
    has_dna_alphabet = False
from snapgene_reader import snapgene_file_to_seqrecord


def set_record_topology(record, topology):
    """Set the Biopython record's topology, possibly passing if already set.

    This actually sets the ``record.annotations['topology']``.The ``topology``
    parameter can be "circular", "linear", "default_to_circular" (will default
    to circular if ``annotations['topology']`` is not already set) or
    "default_to_linear".
    """
    valid_topologies = [
        "circular",
        "linear",
        "default_to_circular",
        "default_to_linear",
    ]
    if topology not in valid_topologies:
        raise ValueError(
            "topology should be one of %s (was %s)."
            % (", ".join(valid_topologies), topology)
        )
    annotations = record.annotations
    default_prefix = "default_to_"
    if topology.startswith(default_prefix):
        if "topology" not in annotations:
            annotations["topology"] = topology[len(default_prefix) :]
    else:
        annotations["topology"] = topology


def record_is_linear(record, default=True):
    """Return true if record.annotations['topology'] == 'linear'"""
    if "topology" not in record.annotations:
        return default
    else:
        return record.annotations["topology"] == "linear"


def load_record(
    record_file,
    topology="default_linear",
    id="auto",
    upperize=True,
    max_name_length=20,
    file_format=None,
):
    """Read a record (from many different input formats).

    Parameters
    ----------
    record_file
        A genbank file, a fasta file, a snapgene file, or a filelike object
        (at which case the format, genbank or fasta, must be given with
        ``file_format``).
    topology
        One of "circular", "linear" or "default_to_circular" (will default
        to circular if ``annotations['topology']`` is not already set).
    id
        Will be used for the record ID and name. If auto, the record id will
        be unchanged unless it is ".", " ", etc. at which case it will be
        replaced by the file name.
    upperize
        If True, the sequence will get upperized.
    max_name_length
        The name of the record will be truncated if too long to avoid Biopython
        exceptions being raised.
    file_format
        Indicates the file format for the parser, when record_file is a filelike
        object.
    """
    if file_format is not None:
        record = SeqIO.read(record_file, file_format)
    elif record_file.lower().endswith(("gb", "gbk")):
        record = SeqIO.read(record_file, "genbank")
    elif record_file.lower().endswith(("fa", "fasta")):
        record = SeqIO.read(record_file, "fasta")
    elif record_file.lower().endswith(".dna"):
        record = snapgene_file_to_seqrecord(record_file)
    else:
        raise ValueError("Unknown format for file: %s" % record_file)
    if upperize:
        record = record.upper()
    set_record_topology(record, topology)
    if id == "auto":
        id = record.id
        if id in [None, "", "<unknown id>", ".", " "]:
            id = os.path.splitext(os.path.basename(record_file))[0]
            record.name = id.replace(" ", "_")[:max_name_length]
        record.id = id
    elif id is not None:
        record.id = id
        record.name = id.replace(" ", "_")[:max_name_length]

    return record


def sequence_to_biopython_record(
    sequence, id="<unknown id>", name="<unknown name>", features=()
):
    """Return a SeqRecord of the sequence, ready to be Genbanked."""
    if has_dna_alphabet:
        seq = Seq(sequence, alphabet=DNAAlphabet())
    else:
        seq = Seq(sequence)

    return SeqRecord(
        seq=seq,
        id=id,
        name=name,
        features=list(features),
        annotations={"molecule_type": "DNA"},
    )


def updated_dict(dic1, dic2):
    """Return dic1 updated with dic2 if dic2 is not None.

    Example
    -------
    >>> my_dict = updated_dict({"size": 7, "color": "r"}, some_defaults_dict)
    """
    if dic2 is not None:
        dic1.update(dic2)
    return dic1


def all_subsets(mylist):
    return itertools.chain.from_iterable(
        itertools.combinations(mylist, k) for k in range(len(mylist) + 1)
    )


def max_min_distance(values_1, values_2, zone=None):
    """Return the maximum distance between one value in one set and its
    nearest neighbor in the other set.

    A large max-min-distance means that the sets values_1 and values_2 are
    visually different, i.e. that one value in one of these sets has no close
    equivalent in the other set.

    Parameters
    ----------
    values_1, values2
        Two lists of values.
     zone
        The max distance will be computed only on couples (v1, v2) where v1 is
        in the zone. Allows to remove border effects.
    """
    values_1, values_2 = np.array(values_1), np.array(values_2)
    all_distances = abs(values_1.reshape((len(values_1), 1)) - values_2)
    if zone is not None:
        mini, maxi = zone
        m1 = all_distances.min(axis=0)[(mini <= values_2) & (values_2 <= maxi)]
        m1 = m1.max() if len(m1) else 0
        m2 = all_distances.min(axis=1)[(mini <= values_1) & (values_1 <= maxi)]
        m2 = m2.max() if len(m2) else 0
        return max(m1, m2)

    return max([all_distances.min(axis=i).max() for i in (0, 1)])
